﻿#region About
// Course material: C# for numerical methods
// Copyright (c) 2009 - present, Igor Grešovnik
#endregion About


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace NumLib
{
    public class EquationSolving
    {

        /// <summary>Example where equation is solved by a static method for finding zeros of a function.</summary>
        /// <remarks><para>Equation solved is defined by f(x) = 0, where the function f(x) is represented by the 
        /// <see cref="EquationSolving.ExampleFunction"/> method and its derivative by the <see cref="EquationSolving.ExampleFunctionDerivative"/> method.</para>
        /// <para>Static method <see cref="EquationSolving.SolveNewtonSimple"/> is used to solve the equation.</para></remarks>
        public static void ExampleStatic()
        {
            RealFunction function = EquationSolving.ExampleFunction;
            RealFunction functionDerivative = EquationSolving.ExampleFunctionDerivative;
            double initialGuess = -5;
            double functionValueTolerance = 1.0e-8;
            double maxStepSize = 1000.0;
            int maxIterations = 50;
            int outputLevel = 1;
            double functionZero;
            double functionValue;
            int numIterations;
            bool successful;
            Console.WriteLine(Environment.NewLine
                + "Solving equation by the newton method, implementation by a static function...");

            // Call the solver method to find function zero:
            SolveNewtonSimple(function, functionDerivative,
                initialGuess, functionValueTolerance,
                maxIterations, maxStepSize, outputLevel,
                out functionZero, out functionValue,
                out numIterations, out successful);

            Console.WriteLine(Environment.NewLine + "Solution procedure finished."
                + Environment.NewLine + "Results: "
                + Environment.NewLine + "                 Successfull: " + (successful ? "Yes." : "No.")
                + Environment.NewLine + "               Function zero: " + functionZero
                + Environment.NewLine + "  Function value in solution: " + functionValue
                + Environment.NewLine + "           Test of the above: " + function(functionZero)
                + Environment.NewLine + "  Number of iterations spent: " + numIterations);
        }


        /// <summary>Example where equation is solved by a solver class for finding zeros of a function.</summary>
        /// <remarks><para>Equation solved is defined by f(x) = 0, where the function f(x) is represented by the 
        /// <see cref="EquationSolving.ExampleFunction"/> method and its derivative by the <see cref="EquationSolving.ExampleFunctionDerivative"/> method.</para>
        /// <para>Class <see cref="EquationSolverNewton"/> is used to solve the equation.</para></remarks>
        public static void ExampleObjectOriented()
        {

            EquationSolverBase.ExampleNewton(EquationSolving.ExampleFunction, EquationSolving.ExampleFunctionDerivative);
        }


        /// <summary>Example function whose zero is searched for in the example.
        /// <para>f(x) = exp(x) - 10</para>
        /// <para>Equation to be solved: exp(x) == 10</para></summary>
        public static double ExampleFunction(double x)
        {
            return (Math.Exp(x) - 10);
        }

        /// <summary>Derivative of the example function whose zero is searched for in the example.
        /// <para>f(x) = exp(x) - 10</para>
        /// <para>f'(x) = exp(x)</para>
        /// <para>Equation to be solved: exp(x) == 10</para></summary>
        public static double ExampleFunctionDerivative(double x)
        {
            return Math.Exp(x);
        }


        /// <summary>Finds function zero by the Newton's method.
        /// <para>This method has some output arguments where the approximate function zero and other
        /// information about the solution are written when the method call completes.</para></summary>
        /// <param name="function">Functio nwhose zero is searched for.</param>
        /// <param name="derivative">Derivative of the function whose zero is searched for.</param>
        /// <param name="initialGuess">Initial guess.</param>
        /// <param name="valueTolerance">Tolerance on absolute value of the function in solution.
        /// Absolute value of the function must be below this value in order for some guess on function 
        /// argument to be recognized as solution (i.e. zero of the function.</param>
        /// <param name="maxIterations">Maximal number of iterations.</param>
        /// <param name="maxStep">Maximal step size.</param>
        /// <param name="outputLevel">Level of output to the console (0 means no output).</param>
        /// <param name="functionZero">Solution is written here.</param>
        /// <param name="functionValue">Function value in the solution is written here.</param>
        /// <param name="numIterations">Number of iterations that were actually spent to find the solution is
        /// written here.</param>
        /// <param name="successful">Whether zero was successfully found within the prescribed tolerance.</param>
        public static void SolveNewtonSimple(RealFunction function, RealFunction derivative, 
            double initialGuess, double valueTolerance, 
            int maxIterations, double maxStep, int outputLevel,
            out double functionZero, out double functionValue,
            out int numIterations, out bool successful)
        {
            double stepReductionFactor = 4.0;
            int maxStepReductions = 20;
            // Check and correct parameters:
            if (function == null)
                throw new ArgumentException("Function whose zero is searched for is not specified (null argument).");
            if (derivative == null)
                throw new ArgumentException("Derivative of the function whose zero is searched for is not specified.");
            if (valueTolerance <= 0)
                valueTolerance = 1.0e-8;
            if (maxIterations < 1)
                maxIterations = 100;
            if (maxStep <= 0)
                maxStep = 1.0e20;
            if (outputLevel > 0)
                Console.WriteLine(Environment.NewLine
                    + "Searching for function zero by the Newton's method. " + Environment.NewLine
                    + "  Initial guess: " + initialGuess + Environment.NewLine
                    + "  Tolerance on function absolute value: " + valueTolerance + Environment.NewLine
                    + "  Maximal number of iterations: " + maxIterations + Environment.NewLine
                    + "  Step limit: " + maxStep);

            // Initialize the loop:
            numIterations = 0;
            bool doStop = false;
            double x = initialGuess;
            double f = function(initialGuess);
            double der = derivative(initialGuess);
            functionZero = x;
            functionValue = f;
            numIterations = 0; 
            successful = false;
            if (Math.Abs(f) < valueTolerance)
            {
                doStop = true;
                successful = true;
            }
            while (!doStop)
            {
                ++numIterations;
                // Calculate the step by the Newton's formula:
                double step = -f / der;
                // Impose the step limit:
                if (Math.Abs(step) > maxStep)
                {
                    if (step < 0)
                        step = -maxStep;
                    else
                        step = maxStep;
                }
                // Calculate the new (trial) guess:
                double xNew = x + step;
                double fNew = function(xNew);
                int numStepReductions = 0;
                while (Math.Abs(fNew) >= Math.Abs(f) && numStepReductions < maxStepReductions)
                {
                    step /= stepReductionFactor;
                    xNew = x + step;
                    fNew = function(xNew);
                    ++numStepReductions;
                    if (outputLevel > 0)
                        Console.WriteLine("  Iteration " + numIterations + ": step reduction No. " + numStepReductions
                            + Environment.NewLine + "    step = " + step
                            + Environment.NewLine + "       x = " + xNew
                            + Environment.NewLine + "    f(x) = " + fNew);
                }
                double derNew = derivative(xNew);
                if (Math.Abs(fNew) < Math.Abs(f))
                {
                    // We obtained a better guess, write it to the output:
                    functionZero = x;
                    functionValue = f;
                }
                x = xNew;
                f = fNew;
                der = derNew;
                // Check for stopping conditions:
                if (Math.Abs(functionValue) < valueTolerance)
                {
                    // Conditions fulfilled, function zero was found successfully:
                    doStop = true;
                    successful = true;
                } else
                {
                    if (numIterations >= maxIterations)
                        doStop = true;
                }
                if (outputLevel > 0)
                {
                    Console.WriteLine("Iteration No. " + numIterations + ": "
                        + Environment.NewLine + "      x = " + x
                        + Environment.NewLine + "   f(x) = " + f
                        + Environment.NewLine + "  f'(x) = " + der
                        + Environment.NewLine + "   step = " + step
                        );
                }
                if (doStop && outputLevel > 0)
                {
                    if (successful)
                    {
                        Console.WriteLine(
                            Environment.NewLine + "Function zero found successfully."
                            + Environment.NewLine + "  Number of iterations: " + numIterations
                            + Environment.NewLine + "              Solution: " + functionZero
                            + Environment.NewLine + "        Function value: " + functionValue
                            );
                    }
                }
            }
        }  // SolveNewtonSimple(...)


    }  // class EquationSolving
}
